# -*- coding: utf-8 -*-
import os

import pytest

PLUGIN_NAME = 'init-check'
INIT_FORBIDDEN_INI = 'init_forbidden_in'
INIT_FILENAME = '__init__.py'
DEFAULT_INIT_FORBIDDEN_IN = {
    'scripts',
    'notebooks',
}


def pytest_addoption(parser):
    group = parser.getgroup('general')
    group.addoption(
        f'--{PLUGIN_NAME}',
        dest=PLUGIN_NAME,
        action='store_true',
        help='Perform checks for __init__.py presence in certain forbidden folders')

    parser.addini(INIT_FORBIDDEN_INI, type='linelist', help=(
        'Each line specifies a single directory name, under which __init__.py files are forbidden'
    ))


def pytest_sessionstart(session):
    """
    Checks whether there are names of directories, where __init__.py is forbidden,
    given in pytest.ini.
    In the case of lack of the forbidden imports specification, error is raised.
    """
    config = session.config
    if config.getoption(PLUGIN_NAME):
        ini_forbidden_inits = config.getini(INIT_FORBIDDEN_INI)
        if ini_forbidden_inits:
            config.ini_forbidden_inits = ini_forbidden_inits
        else:
            config.ini_forbidden_inits = DEFAULT_INIT_FORBIDDEN_IN


def pytest_collect_file(path, parent):
    config = parent.config
    basename = os.path.basename(path.dirname)
    filename = path.basename
    if config.getoption(PLUGIN_NAME) and basename in config.ini_forbidden_inits:
        return InitCheckItem(path, parent, basename, filename)


class ForbiddenInitError(Exception):
    """ Indicates missing newline at the end of file. """

    def __init__(self, path, basename):
        self.path = path.dirname
        self.basename = basename

    def get_message(self):
        return f'{self.path}: {INIT_FILENAME} forbidden under {self.basename} directory'


class InitCheckItem(pytest.Item, pytest.File):

    def __init__(self, path, parent, basename, filename):
        super().__init__(path, parent)
        self.basename = basename
        self.filename = filename

        self.add_marker(PLUGIN_NAME)

    def runtest(self):
        if self.filename == INIT_FILENAME:
            raise ForbiddenInitError(self.fspath, self.basename)

    def repr_failure(self, excinfo):
        if excinfo.errisinstance(ForbiddenInitError):
            return excinfo.value.get_message()
        return super().repr_failure(excinfo)

    def reportinfo(self):
        return self.fspath, -1, PLUGIN_NAME



# -*- coding: utf-8 -*-
"""
NOTE:
    This plugin should have been based on pytest's test discovery, but since this requires
    analysis of gathered tests at different pytest stage, for now plugin implements own
    discovery, which reports failure each time when function / class / file
    contains 'test' token without being under tests/ directory
"""
import os
import re

import pytest

TEST_FILE_PATTERNS = [
    r'test[\w]*\.py',
    r'[\w]*test\.py',
]

PLUGIN_NAME = 'testdir-check'
TESTS_REQUIRED_TOKEN_INI = 'tests_required_token'
DEFAULT_TEST_REQUIRED_TOKEN = {
    'tests',
}


def pytest_addoption(parser):
    group = parser.getgroup('general')
    group.addoption(
        f'--{PLUGIN_NAME}',
        dest=PLUGIN_NAME,
        action='store_true',
        help='Perform initial tests placement checks')

    parser.addini(TESTS_REQUIRED_TOKEN_INI, type='linelist', help=(
        'Each line specifies a single directory name, under which test files are allowed'
    ))


def pytest_sessionstart(session):
    """
    # TODO:
    """
    config = session.config
    if config.getoption(PLUGIN_NAME):
        test_required_token = config.getini(TESTS_REQUIRED_TOKEN_INI)
        if test_required_token:
            config.test_required_token = test_required_token
        else:
            config.test_required_token = DEFAULT_TEST_REQUIRED_TOKEN


def pytest_collect_file(path, parent):
    config = parent.config
    filename = path.basename
    if config.getoption(PLUGIN_NAME) and any(
            re.match(pattern, filename) for pattern in config.test_required_token):
        rel_path = config.rootdir.bestrelpath(path)
        return TestdirCheckItem(path, parent, rel_path, config.test_required_token)


class TestdirCheckError(Exception):
    """ Indicates error during checks of placement of test functions . """

    def __init__(self, path, err_message):
        self.path = path
        self.err_message = err_message

    def get_message(self):
        return f'{self.path}: {self.err_message}'


class TestdirCheckItem(pytest.Item, pytest.File):

    def __init__(self, path, parent, rel_path, test_required_token):
        super().__init__(path, parent)
        self.rel_path = rel_path
        self.test_required_token = test_required_token
        self.add_marker(PLUGIN_NAME)

    def runtest(self):
        testdir_err = check_testdir(self.rel_path, self.test_required_token)

        if testdir_err:
            raise TestdirCheckError(self.fspath, testdir_err)

    def repr_failure(self, excinfo):
        if excinfo.errisinstance(TestdirCheckError):
            return excinfo.value.get_message()
        return super().repr_failure(excinfo)

    def reportinfo(self):
        return self.fspath, -1, PLUGIN_NAME


def check_testdir(rel_path, test_required_token):
    path_components = rel_path.split(os.sep)
    if any(test_required_token == component for component in path_components):
        return ''
    return f'Invalid test file placement'



# -*- coding: utf-8 -*-
import json

import pytest

PLUGIN_NAME = 'notebook-check'
NOTEBOOK_ALLOWED_INI = 'notebook_allowed_in'
JUPYTER_NOTEBOOK = '.ipynb'
DEFAULT_NOTEBOOK_ALLOWED = {
    'notebooks',
}


def pytest_addoption(parser):
    group = parser.getgroup('general')
    group.addoption(
        f'--{PLUGIN_NAME}',
        dest=PLUGIN_NAME,
        action='store_true',
        help='Perform checks whether notebooks has been cleaned'
             ' and are placed under proper (specified) directory')

    parser.addini(NOTEBOOK_ALLOWED_INI, type='linelist', help=(
        'Each line specifies a single directory name, under which jupyter notebooks might be placed'
    ))


def pytest_sessionstart(session):
    """
    Checks whether in pytest.ini directory names, where jupyter notebooks are allowed, was provided
    In the case of lack of the notebooks directory names specification, error is raised.
    """
    config = session.config
    if config.getoption(PLUGIN_NAME):
        ini_allowed_notebook_dirs = config.getini(NOTEBOOK_ALLOWED_INI)
        if ini_allowed_notebook_dirs:
            config.ini_allowed_notebook_dirs = ini_allowed_notebook_dirs
        else:
            config.ini_allowed_notebook_dirs = DEFAULT_NOTEBOOK_ALLOWED


def pytest_collect_file(path, parent):
    config = parent.config
    if config.getoption(PLUGIN_NAME) and path.ext == JUPYTER_NOTEBOOK:
        return NotebookCheckItem(path, parent)


def check_notebook_dir(path, allowed_dirs):
    if not any(name in path.strpath for name in allowed_dirs):
        return f'notebook not under one of allowed directories: {allowed_dirs}'
    return ''


def check_notebook_output(path):
    with open(path, 'r') as notebook_file:
        notebook = json.load(notebook_file)
        for cell in notebook['cells']:
            if (cell.get('ExecuteTime') is not None or
                    cell.get('execution_count') is not None or
                    cell.get('outputs')):
                return 'notebook with output'
    return ''


class NotebookCheckError(Exception):
    """ Indicates incorrect .ipnyb file placement. """

    def __init__(self, path, error_messages):
        self.path = path
        self.error_messages = error_messages

    def get_message(self):
        return f'{self.path}: {", ".join(self.error_messages)}\n'


class NotebookCheckItem(pytest.Item, pytest.File):

    def __init__(self, path, parent):
        super().__init__(path, parent)
        self.add_marker(PLUGIN_NAME)

    def runtest(self):
        error_messages = []
        dir_error = check_notebook_dir(self.fspath, self.config.ini_allowed_notebook_dirs)
        if dir_error:
            error_messages.append(dir_error)

        output_error = check_notebook_output(self.fspath)
        if output_error:
            error_messages.append(output_error)

        if error_messages:
            raise NotebookCheckError(str(self.fspath), error_messages)

    def repr_failure(self, excinfo):
        if excinfo.errisinstance(NotebookCheckError):
            return excinfo.value.get_message()
        return super().repr_failure(excinfo)

    def reportinfo(self):
        return self.fspath, -1, PLUGIN_NAME



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os

from setuptools import setup

import pytest_alphamoon


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding='utf-8').read()


setup(
    name='pytest-alphamoon',
    version=pytest_alphamoon.__version__,
    author='Jakub Binkowski',
    author_email='jakub.binkowski@alphamoon.ai',
    maintainer='Jakub Binkowski',
    maintainer_email='dev@alphamoon.ai',
    license='MIT',
    url='https://gitlab.com/alphamoon/internal_tools/pytest_alphamoon',
    description='Static code checks used at Alphamoon',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    py_modules=['pytest_alphamoon'],
    python_requires='>=3.6',
    install_requires=['pytest>=3.5.0'],
    packages=['pytest_alphamoon'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Pytest',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Testing',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
    ],
    entry_points={
        'pytest11': [
            'init_check = pytest_alphamoon.init_check',
            'import_check = pytest_alphamoon.import_check',
            'newline_check = pytest_alphamoon.newline_check',
            'notebook_check = pytest_alphamoon.notebook_check',
            'testdir_check = pytest_alphamoon.testdir_check',
        ],
    },
)



# -*- coding: utf-8 -*-
import ast

import pytest

PLUGIN_NAME = 'import-check'
FORBIDDEN_IMPORT_INI = 'import_forbidden_from'
DEFAULT_IMPORTS_FORBIDDEN_FROM = {
    'scripts',
    'user_settings',
    'notebooks',
}


def pytest_addoption(parser):
    group = parser.getgroup('general')
    group.addoption(
        f'--{PLUGIN_NAME}',
        dest=PLUGIN_NAME,
        action='store_true',
        help='Perform initial import checks')

    parser.addini(FORBIDDEN_IMPORT_INI, type='linelist', help=(
        'Each line specifies a single name (e.g. module name), from which imports are forbidden'
    ))


def pytest_sessionstart(session):
    """
    Checks whether there are names of imports given in pytest.ini.
    In the case of lack of the forbidden imports specification, error is raised.
    """
    config = session.config
    if config.getoption(PLUGIN_NAME):
        ini_forbidden_imports = config.getini(FORBIDDEN_IMPORT_INI)
        if ini_forbidden_imports:
            config.ini_forbidden_imports = ini_forbidden_imports
        else:
            config.ini_forbidden_imports = DEFAULT_IMPORTS_FORBIDDEN_FROM


def pytest_collect_file(path, parent):
    config = parent.config
    if config.getoption(PLUGIN_NAME) and path.ext == '.py':
        return ImportCheckItem(path, parent)


class ImportCheckError(Exception):
    """ Indicates error during import checks. """

    def __init__(self, path, forbidden_imports):
        self.path = path
        self.forbidden_imports = forbidden_imports

    def get_message(self):
        msg = ''
        for line_num, import_type in self.forbidden_imports:
            msg += f'{self.path}:{line_num}: imports from {import_type} are forbidden\n'
        return msg


class ImportCheckItem(pytest.Item, pytest.File):

    def __init__(self, path, parent):
        super().__init__(path, parent)
        self.add_marker(PLUGIN_NAME)

    def runtest(self):
        checker = ImportChecker(self.fspath, self.config.ini_forbidden_imports)
        forbidden_imports = checker.check()
        if forbidden_imports:
            raise ImportCheckError(str(self.fspath), forbidden_imports)

    def repr_failure(self, excinfo):
        if excinfo.errisinstance(ImportCheckError):
            return excinfo.value.get_message()
        return super().repr_failure(excinfo)

    def reportinfo(self):
        return self.fspath, -1, PLUGIN_NAME


class ImportChecker(ast.NodeVisitor):
    def __init__(self, path, forbidden_imports):
        self.path = path
        self.violating_imports = []
        self.forbidden_imports = forbidden_imports

    def check(self):
        self.violating_imports = []
        with open(self.path, 'r') as src_code:
            ast_tree = ast.parse(src_code.read())
            self.visit(ast_tree)
            return self.violating_imports

    def visit_Import(self, node):
        for stmt in node.names:
            self._check_import_forbidden(stmt.name, node.lineno)
        self.generic_visit(node)

    def visit_ImportFrom(self, node):
        self._check_import_forbidden(node.module, node.lineno)
        for stmt in node.names:
            self._check_import_forbidden(stmt.name, node.lineno)
        self.generic_visit(node)

    def _check_import_forbidden(self, import_chunk, line_num):
        for forbid_import in self.forbidden_imports:
            if forbid_import in import_chunk.split('.'):
                self.violating_imports.append((line_num, forbid_import))



# -*- coding: utf-8 -*-
import os

import pytest

PLUGIN_NAME = 'newline-check'
NEWLINE_REQUIRE_INI = 'newline_require'
DEFAULT_REQUIRED_NEWLINE_IN = {
    'requirements*',
    '.gitignore',
}


def pytest_addoption(parser):
    group = parser.getgroup('general')
    group.addoption(
        f'--{PLUGIN_NAME}',
        dest=PLUGIN_NAME,
        action='store_true',
        help='Perform checks on specified files whether there is newline present'
             ' at the end of file')

    parser.addini('newline_require', type='linelist', help=(
        'Each line specifies a glob filename patter which will be checked against newline. '
        'Example: */requirements.txt'
    ))


def pytest_sessionstart(session):
    """
    Checks whether there are filename patters to be checked against newline, as specified by user
    If user specifies filenames, they overrides default ones!
    """
    config = session.config
    if config.getoption(PLUGIN_NAME):
        ini_newline_require = config.getini(NEWLINE_REQUIRE_INI)
        if ini_newline_require:
            config.require_newline = FileMatcher(ini_newline_require)
        else:
            config.require_newline = FileMatcher(DEFAULT_REQUIRED_NEWLINE_IN)


def pytest_collect_file(path, parent):
    config = parent.config
    if config.getoption(PLUGIN_NAME) and config.require_newline(path):
        return NewlineCheckItem(path, parent)


def check_newline(path):
    """
    Performs newline checks and returns rule violation status. If file is empty, test is passed.
    Otherwise file is checked against occurrence of newline at the end and eventually
    against forbidden multiple newlines at the end
    """
    with open(path, 'r') as f:
        lines = f.readlines()
        if not lines:
            return False
        if lines[-1] == '\n':
            return 'Extra newline(s) at the end of file'
        if not lines[-1].endswith('\n'):
            return 'Missing newline at the end of file'
        return False


class FileMatcher:
    """
    Class helps to maintain custom list of files, which are forced to end with newline character
    Based on pytest-isort implementation:
        https://github.com/moccu/pytest-isort/blob/master/pytest_isort.py
    """

    def __init__(self, require_lines):
        self.requires = []

        for line in require_lines:
            comment_position = line.find("#")
            # Strip comments.
            if comment_position != -1:
                line = line[:comment_position]

            glob = line.strip()

            # Skip blank lines.
            if not glob:
                continue

            # Normalize path if needed.
            if glob and os.sep != '/' and '/' in glob:
                glob = glob.replace('/', os.sep)

            self.requires.append(glob)

    def __call__(self, path):
        for glob in self.requires:
            if path.fnmatch(glob):
                return glob


class MissingNewlineError(Exception):
    """ Indicates missing newline at the end of file. """

    def __init__(self, path, message):
        self.path = path
        self.message = message

    def get_message(self):
        return f'{self.path}: {self.message}'


class NewlineCheckItem(pytest.Item, pytest.File):

    def __init__(self, path, parent):
        super().__init__(path, parent)
        self.add_marker(PLUGIN_NAME)

    def runtest(self):
        newline_err = check_newline(self.fspath)

        if newline_err:
            raise MissingNewlineError(str(self.fspath), newline_err)

    def repr_failure(self, excinfo):
        if excinfo.errisinstance(MissingNewlineError):
            return excinfo.value.get_message()
        return super().repr_failure(excinfo)

    def reportinfo(self):
        return self.fspath, -1, PLUGIN_NAME
